# ----------------------------------------------------------------------------
# Copyright (c) 2021, Diego Garcia Huerta.
#
# Your use of this software as distributed in this GitHub repository, is
# governed by the MIT License
#
# Your use of the Shotgun Pipeline Toolkit is governed by the applicable
# license agreement between you and Autodesk / Shotgun.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os

import sgtk

import GafferScene
import GafferImage

__author__ = "Diego Garcia Huerta"
__contact__ = "https://www.linkedin.com/in/diegogh/"


HookBaseClass = sgtk.get_hook_baseclass()


# let's put some color to these gray-ish UIs
ITEM_COLORS = {"scene_reader": "#e7a81d", "image_reader": "#a8e71d"}


def nodes_of_type(node_type, node=None, result=None):
    if result is None:
        result = []

    if isinstance(node, node_type):
        result.append(node)

    for child in node.children():
        nodes_of_type(node_type, node=child, result=result)

    return result


class BreakdownSceneItem(str):
    """
    Helper Class to store metadata per update item.
    tk-multi-breakdown requires item['node'] to be a str. This is what is displayed in
    the list of recognized items to update. We want to add metadata to each item
    as what we want to display as name is not the actual item to update.
    As a str is required we are forced to inherit from str instead of the more
    python friendly object + __repr__ magic method.
    """

    def __new__(cls, node, node_type):
        node_name = node.getName()
        node_filename = os.path.basename(node["fileName"].getValue())

        text = (
            "<span style='color:%s'><b>%s</b></span>"
            "<br/><nobr>%s</nobr>"
            "<br/><nobr><b><sub>%s</sub></b></nobr>"
            % (ITEM_COLORS[node_type], node_name, node_filename, node_type)
        )

        item = str.__new__(cls, text)
        item.node = node
        item.node_type = node_type

        return item


class BreakdownSceneOperations(HookBaseClass):
    """
    Breakdown operations for Natron.

    This implementation handles detection of natron read and write nodes.
    """

    def scan_scene(self):
        """
        The scan scene method is executed once at startup and its purpose is
        to analyze the current scene and return a list of references that are
        to be potentially operated on.

        The return data structure is a list of dictionaries. Each scene
        reference that is returned should be represented by a dictionary with
        three keys:

        - "attr": The filename attribute of the 'node' that is to be operated
           on. Most DCCs have a concept of a node, attribute, path or some
           other way to address a particular object in the scene.
        - "type": The object type that this is. This is later passed to the
           update method so that it knows how to handle the object.
        - "path": Path on disk to the referenced object.

        Toolkit will scan the list of items, see if any of the objects matches
        any templates and try to determine if there is a more recent version
        available. Any such versions are then displayed in the UI as out of
        date.
        """

        # Introspect the natron scene for read and write nodes
        # so we can gather the filenames available.

        refs = []

        engine = sgtk.platform.current_engine()

        scene_reader_nodes = nodes_of_type(GafferScene.SceneReader, node=engine.script)
        image_reader_nodes = nodes_of_type(GafferImage.ImageReader, node=engine.script)

        for node in scene_reader_nodes:
            ref_path = node["fileName"].getValue()
            ref_path = ref_path.replace("/", os.path.sep)
            refs.append(
                {
                    "node": BreakdownSceneItem(node, "scene_reader"),
                    "type": "scene_reader",
                    "path": ref_path,
                }
            )

        for node in image_reader_nodes:
            ref_path = node["fileName"].getValue()
            ref_path = ref_path.replace("/", os.path.sep)
            refs.append(
                {
                    "node": BreakdownSceneItem(node, "image_reader"),
                    "type": "image_reader",
                    "path": ref_path,
                }
            )

        return refs

    def update(self, items):
        """
        Perform replacements given a number of scene items passed from the app.

        Once a selection has been performed in the main UI and the user clicks
        the update button, this method is called.

        The items parameter is a list of dictionaries on the same form as was
        generated by the scan_scene hook above. The path key now holds
        the that each attribute should be updated *to* rather than the current
        path.
        """

        for i in items:
            new_path = i["path"]
            node = i["node"].node

            node["fileName"].setValue(new_path)
            node["refreshCount"].setValue(node["refreshCount"].getValue() + 1)
